from collections import Iterable
from typing import Any


def is_iterable(d: Any, exclude_str: bool = True) -> bool:
    """
    refer url: https://docs.python.org/3/glossary.html#term-iterator
    An object capable of returning its members one at a time.
    Examples of iterables include all sequence types (such as list, str, and tuple)
    and some non-sequence types like dict, file objects,
    and objects of any classes you define with an __iter__() method or with a __getitem__() method
    that implements Sequence semantics.
    :param d:
    :return:
    """
    if exclude_str:
        return isinstance(d, Iterable) and not isinstance(d, str)
    else:
        return isinstance(d, Iterable)



# -*- coding: utf-8 -*-

import json
from datetime import date
from datetime import datetime
from decimal import Decimal
from typing import Any, Union


class ComplexEncoder(json.JSONEncoder):
    """
    usage:
    json.dumps({'now':now}, cls=ComplexEncoder)

    """

    def default(self, obj: Union[date, datetime]) -> Any:
        if isinstance(obj, datetime):
            return obj.strftime('%Y-%m-%d %H:%M:%S')
        elif isinstance(obj, date):
            return obj.strftime('%Y-%m-%d')
        elif isinstance(obj, Decimal):
            return float(obj)
        elif getattr(obj, '__name__', ''):
            return obj.__name__
        else:
            return str(obj)


def dumps(py_ins: Any) -> str:
    u"""
    把 py_ins 转化为 json 字符串对象;
    这个封装方法的优势：
    + sort_keys=True，避免 dict 对象转化为 json str 时因位置不同导致 json 不同；
    + 支持 datatime 的转化；
    :param py_ins:
    :return:
    """
    return json.dumps(py_ins, sort_keys=True, cls=ComplexEncoder)



# -*- coding: utf-8 -*-

from typing import Dict
from typing import Tuple


class Phone:
    f_yidong = 'yidong'
    f_liantong = 'liantong'
    f_dianxin = 'dianxin'
    valid_length = 11

    @classmethod
    def get_operator_segment(cls) -> Dict[str, Tuple]:
        return {
            cls.f_yidong: ('134', '135', '136', '137', '138', '139', '150', '151', '152', '157',
                           '158', '159', '182', '183', '184', '187', '188', '147', '178', '1703', '1705', '1706'),
            cls.f_liantong: ('130', '131', '132', '145', '155', '156', '175', '176', '185', '186', '171', '1704',
                             '1707', '1708', '1709'),
            cls.f_dianxin: ('133', '149', '153', '173', '177', '180', '181', '189', '1700', '1701', '1702')
        }

    @classmethod
    def all_operators(cls) -> Tuple[str, str, str]:
        return (cls.f_yidong, cls.f_liantong, cls.f_dianxin)

    @classmethod
    def get_operator(cls, phone: str) -> str:
        if not isinstance(phone, str):
            raise ValueError(f"phone {phone} is expected str and got {type(phone)}")
        if len(phone) != cls.valid_length:
            raise ValueError(f"phone {phone} length is expected {cls.valid_length} and got {len(phone)}")
        top_three = phone[0:3]
        top_four = phone[0:4]
        for operator, segment in cls.get_operator_segment().items():
            if top_three in segment or top_four in segment:
                return operator
        return ''



from typing import Dict, Any

import pandas as pd


class DataFrame:

    def __new__(self, data: Dict[str, Any]) -> pd.DataFrame:
        if isinstance(data, dict):
            return pd.DataFrame(data, index=[0])
        else:
            raise NotImplementedError()


import os
from typing import Tuple, Optional, List, Mapping

import requests


def apply_async(host: str,
                api: str,
                auth: Tuple = None,
                args: Optional[List] = None,
                kwargs: Optional[Mapping] = None,
                queue: str = ''):
    is_valid = lambda x: x[0] in ('args', 'kwargs', 'queue') and x[1]
    url = os.path.join(host, 'api/task/async-apply', api)
    parameter = dict(filter(is_valid, locals().items()))
    send_task = requests.post(url, json=parameter, auth=auth) if auth else requests.post(url, json=parameter)
    if not send_task.ok:
        raise SystemError('request url {} failed, flower not work'.format(url))
    return send_task.json()



import operator as op
from collections.abc import Mapping
from datetime import datetime, timedelta
from functools import wraps
from typing import Callable, Optional, List, Dict

from jwt.api_jwt import PyJWT
from jwt.exceptions import ExpiredSignatureError
from nezha.ustring import to_str


def _encode(func: Callable):
    """
    wrap jwt.encode function.
    adding lifetime parameter to jwt.encode.
    if not specify lifetime, default to 3600s.

    :param func:
    :param lifetime: the period of validity for token. the unit of lifetime is second.
    :return:
    """

    @wraps(func)
    def wrap(self, *args, **kwargs):
        if len(args) < 0 or not isinstance(args[0], Mapping):
            raise TypeError('Expecting a mapping object, as JWT only supports JSON objects as payloads.')
        lifetime = kwargs.pop('lifetime', None)
        default_lifetime = 3600
        lifetime_val = lifetime if lifetime and isinstance(lifetime, int) else default_lifetime
        args[0]['exp'] = op.add(datetime.utcnow(), timedelta(seconds=lifetime_val))
        return to_str(func(self, *args, **kwargs))

    return wrap


def _is_expired(self, jwt: str, key: str, verify: bool = True,
               algorithms: Optional[List[str]] = None,
               options: Optional[Dict] = None, **kwargs):
    try:
        self.decode(jwt, key=key, verify=verify, algorithms=algorithms, options=options, **kwargs)
        return False
    except ExpiredSignatureError as e:
        return True


def _is_valid(self, jwt: str, key: str, verify: bool = True,
             algorithms: Optional[List[str]] = None,
             options: Optional[Dict] = None, **kwargs):
    try:
        self.decode(jwt, key=key, verify=verify, algorithms=algorithms, options=options, **kwargs)
        return True
    except Exception as e:
        return False


PyJWT.is_valid = _is_valid
PyJWT.encode = _encode(PyJWT.encode)
PyJWT.is_expired =  _is_expired
_jwt_global_obj = PyJWT()
encode = _jwt_global_obj.encode
is_valid = _jwt_global_obj.is_valid
is_expired = _jwt_global_obj.is_expired



# -*-coding: utf-8-*-

import datetime as dt
import time
from datetime import datetime
from typing import Union

DEFAULT_FMT = '%Y-%m-%d %H:%M:%S'
NO_BLANK_FMT = '%Y%m%d%H%M%S'


def strftime(obj: Union[datetime, dt.date, None] = None, fmt: str = DEFAULT_FMT) -> str:
    """
    if obj, format obj to string.
    else format now time to string.
    :param obj:
    :param fmt:
    :return:
    """
    if obj and isinstance(obj, (datetime, dt.date)):
        return obj.strftime(fmt)
    return time.strftime(fmt)


def timestamp(precision: str = 's', returned_str: bool = False) -> Union[int, str]:
    """

    :param precision: s means seconds, ms means millisecond, others is not support
    :param returned_str: returned type is str or int
    :return:
    """
    origin = time.time()
    may_ms = origin * 1000 if precision == 'ms' else origin
    to_int = int(may_ms)
    return str(to_int) if returned_str else to_int


def _precision(precision: str):
    table = {
        ('second', 'S', 'Second', 's'): DEFAULT_FMT,
        ('M', 'minute', 'Minute'): '%Y-%m-%d %H:%M',
        ('H', 'hour', 'Hour'): '%Y-%m-%d %H',
        ('d', 'day', 'Day'): '%Y-%m-%d'
    }
    return table[tuple(filter(lambda x: precision in x, table.keys()))[0]]


def today(return_str: bool = True, precision: str = 'second') -> Union[str, dt.date]:
    """

    :param return_str:
    :param fmt:
    :return: 2019-08-10 16:23:17
    """
    t = dt.date.today()
    if return_str:
        return strftime(obj=t, fmt=_precision(precision))
    return t


def yesterday(return_str: bool = True, precision: str = 's') -> Union[str, dt.date]:
    y = today(return_str=False) + dt.timedelta(-1)
    if return_str:
        return strftime(obj=y, fmt=_precision(precision))
    return y

def last_month(fmt='%Y-%m'):
    last_day_of_last_month = dt.date.today().replace(day=1) - dt.timedelta(days=1)
    return last_day_of_last_month.strftime(fmt)


if __name__ == '__main__':
    print(last_month())
